@@ -1,3 +1,5 @@
+use std::{borrow::Cow, ops::Add};
+
 use crate::Parsed;
 
 use super::shared::{
@@ -139,7 +141,7 @@ macro_rules! basename {
 }
 pub use basename;
 
-/// Returns the directory name of a path, similar to the Unix dirname command. Trailing directory separators are ignored, 
+/// Returns the directory name of a path, similar to the Unix dirname command. Trailing directory separators are ignored,
 /// ```rust
 /// assert_eq!(&nodejs_path::dirname("/foo/bar/baz/asdf/quux"), "/foo/bar/baz/asdf");
 /// ```
@@ -206,7 +208,6 @@ pub fn extname(path: &str) -> String {
 
 /// Returns a path string from an object. This is the opposite of nodejs_path::parse().
 
-
 pub fn format(path_object: Parsed) -> String {
     format_inner("/", path_object)
 }
@@ -214,22 +215,18 @@ pub fn format(path_object: Parsed) -> String {
 /// The method determines if path is an absolute path. If the given path is a zero-length string, false will be returned.
 /// #Example
 /// ```rust
-/// assert_eq!(nodejs_path::posix::is_absolute("/foo/bar"), true); 
-/// assert_eq!(nodejs_path::posix::is_absolute("/baz/.."), true); 
-/// assert_eq!(nodejs_path::posix::is_absolute("qux/"), false); 
+/// assert_eq!(nodejs_path::posix::is_absolute("/foo/bar"), true);
+/// assert_eq!(nodejs_path::posix::is_absolute("/baz/.."), true);
+/// assert_eq!(nodejs_path::posix::is_absolute("qux/"), false);
 /// assert_eq!(nodejs_path::posix::is_absolute("."), false);  
 /// assert_eq!(nodejs_path::posix::is_absolute(""), false);  
 /// ```
 pub fn is_absolute(path: &str) -> bool {
-    path.chars()
-        .into_iter()
-        .next()
-        .map(|c| c == CHAR_FORWARD_SLASH)
-        .unwrap_or(false)
+    path.bytes().next().map(|c| c == b'/').unwrap_or(false)
 }
 
 /// The method joins all given path segments together using the platform-specific separator as a delimiter, then normalizes the resulting path.
-/// 
+///
 /// Zero-length path segments are ignored. If the joined path string is a zero-length string then '.' will be returned, representing the current working directory.
 /// ```rust
 /// assert_eq!(nodejs_path::posix::join!("/foo", "bar", "baz/asdf", "quux", ".."), "/foo/bar/baz/asdf");
@@ -252,30 +249,34 @@ pub fn join_impl(args: &[&str]) -> String {
     if args.len() == 0 {
         ".".to_owned()
     } else {
-        let mut joined: Option<String> = None;
-        args.iter().for_each(|arg| {
-            if arg.len() > 0 {
-                if let Some(joined) = &mut joined {
-                    joined.push('/');
-                    joined.push_str(*arg);
+        // let length =
+        let joined = args
+            .iter()
+            .filter_map(|&arg| {
+                if arg.is_empty() {
+                    None
                 } else {
-                    joined = Some(arg.to_string())
+                    Some(Cow::Borrowed(arg))
                 }
-            };
-        });
-
-        joined
-            .map(|joined| normalize(&joined))
-            .unwrap_or(".".to_owned())
+            })
+            .reduce(|mut pre, cur| {
+                pre = pre.add("/");
+                pre = pre.add(cur);
+                pre
+            });
+        match joined {
+            Some(joined) => normalize(&joined),
+            None => ".".to_string(),
+        }
     }
 }
 
 /// The path.normalize() method normalizes the given path, resolving '..' and '.' segments.
-/// 
+///
 /// When multiple, sequential path segment separation characters are found (e.g. / on POSIX and either \ or / on Windows), they are replaced by a single instance of the platform-specific path segment /// separator (/ on POSIX and \ on Windows). Trailing separators are preserved.
-/// 
+///
 /// If the path is a zero-length string, '.' is returned, representing the current working directory.
-/// 
+///
 /// ```rust
 /// assert_eq!(nodejs_path::posix::normalize("/foo/bar//baz/asdf/quux/.."), "/foo/bar/baz/asdf");
 /// ```
@@ -289,28 +290,51 @@ pub fn normalize(path: &str) -> String {
             .last()
             .map(|c| c == CHAR_FORWARD_SLASH)
             .unwrap_or(false);
-
-        let mut path = normalize_string(path, !is_absolute, &'/', &is_posix_path_separator);
-
-        if path.len() == 0 {
-            if is_absolute {
-                return "/".to_owned();
-            } else if trailing_separator {
-                return "./".to_owned();
-            } else {
-                return ".".to_owned();
-            }
-        }
-
-        if trailing_separator {
-            path.push('/');
-        }
-
-        if is_absolute {
-            return format!("/{}", path);
+        let mut consecutive_dd = 0;
+        // let mut path = normalize_string(path, !is_absolute, &'/', &is_posix_path_separator);
+        let mut path_stack = vec![];
+        path.split("/")
+            .filter(|seg| !seg.is_empty())
+            .for_each(|seg| {
+                match seg {
+                    "." => {}
+                    ".." => {
+                        // path_stack.pop();
+                        if consecutive_dd == path_stack.len() {
+                            path_stack.push(seg);
+                            consecutive_dd += 1;
+                        } else {
+                            path_stack.pop();
+                        }
+                    }
+                    other => {
+                        path_stack.push(other);
+                    }
+                }
+            });
+        let mut normalized_path = if is_absolute {
+            // if is absolute path, whatever how many times .. used, it is just the same as /
+            path_stack
+                .iter()
+                .position(|&str| str != "..")
+                .map(|item| path_stack[item..].join("/"))
+                .unwrap_or("".to_string())
         } else {
-            return path;
+            path_stack.join("/")
+        };
+        if is_absolute {
+            normalized_path = "/".to_string() + &normalized_path;
         }
+
+        if normalized_path.is_empty() {
+            normalized_path.push('.');
+        }
+
+        if trailing_separator && normalized_path != "/" {
+            normalized_path.push('/');
+        }
+
+        normalized_path
     }
 }
 
@@ -564,7 +588,7 @@ pub fn resolve_impl(args: &[&str]) -> String {
     if resolved_absolute {
         "/".to_owned() + &resolved_path
     } else {
-        if resolved_path.len() > 0 {
+        if !resolved_path.is_empty() {
             resolved_path
         } else {
             ".".to_owned()
